{%- from "crm/widgets/macros.html" import m_render_phonenumber_form %}

<table class="table js-widget" data-init-with="dynamicRowWidget" data-prefix="{{ field.name }}">
  <thead>
    <tr>
        <th></th>
    </tr>
  </thead>
  <tbody>
    {%- for fieldform in field %}
    <tr>
      <td>
        {{ m_render_phonenumber_form(fieldform) }}
      </td>
    </tr>
    {%- endfor %}
  </tbody>
</table>