{%- extends "base.html" %}

{% block content %}

  <h1>{{ _('Generating file') }}</h1>

  <div class="progress progress-bar-striped js-widget"
       data-init-with="crmExcelExport"
       data-task-id="{{ task.id }}"
       data-index-url="{{ g.view.index_url() }}">
    <div class="progress-bar" role="progressbar" aria-valuenow="0" aria-valuemin="0" aria-valuemax="100" style="min-width: 2em; width: 100%;">
      {{ _('Waiting to start') }}
    </div>
  </div>

  <div class="download-btn hide">
    <a class="btn btn-default">
      <i class="fa fa-download"></i>
      <span class="filename"></span>
    </a>
  </div>
{% endblock %}
